#include "paramSimul.h"
#include "ue.h"
#include "objectBSSectorUe.h"
#include "paramFastFading.h"
#include "rvGenerater.h"

void Generate_channel_samples(int ue_idx, int adj_sec_idx);
void Compute_Channel_Coef(MatrixXcd *H, double *P,double K_R, int num_cluster, int num_ray_per_cluster, double **phi_AOA, double **phi_AOD, double **theta_ZOA,double **theta_ZOD, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD, double **kappa, int LOS_condition, double *v);
MatrixXcd compute_H_NLOS(double F_rx_theta, double F_rx_phi, double F_tx_theta, double F_tx_phi, double *P, int n, int m,int M, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double **kappa,  double *v);
MatrixXcd compute_H_LOS(double F_rx_theta, double F_rx_phi, double F_tx_theta, double F_tx_phi, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD, double *v);
void Generate_delays(double K, double DS, double *tau,double *tau_LOS, int num_cluster, int propagation_condition);
int Generate_cluster_powers(double K_R, double P_1LOS, double DS, double *tau, double *P, int num_cluster, int propagation_condition);
void Generate_arrival_departure_angle(double K, double *P,double ASA, double ASD, double ZSA, double ZSD,double mu_offset_ZOD, int num_cluster, int num_ray_per_cluster, int propagation_condition, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double phi_LOS_AOA,double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD);
void Couple_rays();
void Generate_XPR(double mu_XPR, double sigma_XPR, int num_cluster, int num_ray_per_cluster, double **kappa);
double max(double *A, int size);

bool check_delay = false;
bool check_cluster_power = false;
bool check_AOA = false;
bool check_AOD = false;
bool check_ZOA = false;
bool check_ZOD = false;
bool check_XPR = false;

void Generate_SSP()
{
	cout << "Generate_SSP" << endl;
	for (int ue_idx = 0; ue_idx < num_ue; ue_idx++)
	{
		cout << "ue_idx :" << ue_idx << endl;
		for (int adj_sec_idx = 0; adj_sec_idx < num_compute_coef; adj_sec_idx++)
		{
			Generate_channel_samples(ue_idx, adj_sec_idx);
		}
	}
	getchar();

}


/*===================================================================
FUNCTION: Generate_channel_samples()

AUTHOR:

DESCRIPTION: compute channel coefficients and stored

NOTES: Output is a Channel_Coef.

===================================================================*/
void Generate_channel_samples(int ue_idx, int adj_sec_idx)
{
	int bs_idx = adj_sec_idx % 3;
	int propagation_condition = ue[ue_idx].propagation_condition[bs_idx];
	int LOS_condition = ue[ue_idx].LOS_condition[bs_idx];
	int N = num_cluster[propagation_condition];
	int M = num_ray_per_cluster[propagation_condition];
	double v[3] = {ue[ue_idx].v[0],ue[ue_idx].v[1],ue[ue_idx].v[2]};
		
	double phi_LOS_AOA = ue[ue_idx].phi_LOS_AOA[adj_sec_idx];
	double phi_LOS_AOD = ue[ue_idx].phi_LOS_AOD[adj_sec_idx];
	double theta_LOS_ZOA = ue[ue_idx].theta_LOS_ZOA[adj_sec_idx];
	double theta_LOS_ZOD = ue[ue_idx].theta_LOS_ZOD[adj_sec_idx];
	double mu_offset_ZOD = ue[ue_idx].ZOD_offset[bs_idx];

	double DS;
	double K = 0;
	double K_R = 0;
	double P_1LOS = 0;
	double *P;
	double *tau;
	double *tau_LOS;
	double ASA;
	double ASD;
	double ZSA;
	double ZSD;
	double mu_XPR;
	double sigma_XPR;
	double **phi_AOA;
	double **phi_AOD;
	double **theta_ZOA;
	double **theta_ZOD;
	double **kappa;
	MatrixXcd *H;
	double *delay;


	tau = new double[N];
	tau_LOS = new double[N];
	P = new double[N];
	


	phi_AOA = new double *[N];
	phi_AOD = new double *[N];
	theta_ZOA = new double *[N];
	theta_ZOD = new double *[N];
	kappa = new double *[N];

	for (int n = 0; n < N; n++)
	{
		phi_AOA[n] = new double[M];
		phi_AOD[n] = new double[M];
		theta_ZOA[n] = new double[M];
		theta_ZOD[n] = new double[M];
		kappa[n] = new double[M];
	}

	H = new MatrixXcd[N + 4];
	delay = new double[N + 4];

	for (int n=0; n<N+4;n++)
	{
		H[n] = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, P_bs*Mg_bs*Ng_bs*M_bs*N_bs);
	}

	
	switch (propagation_condition)
	{
	case LOS:
		DS = ue[ue_idx].LSPs[bs_idx](LOS_DS);
		K = ue[ue_idx].LSPs[bs_idx](LOS_K);
		ASA = ue[ue_idx].LSPs[bs_idx](LOS_ASA);
		ASD = ue[ue_idx].LSPs[bs_idx](LOS_ASD);
		ZSA = ue[ue_idx].LSPs[bs_idx](LOS_ZSA);
		ZSD = ue[ue_idx].LSPs[bs_idx](LOS_ZSD);
		K_R = pow(10, K / 10);
		P_1LOS = K_R/(K_R+1);
		mu_XPR = XPR_mean[LOS];
		sigma_XPR = XRP_std[LOS];

		break;
	case NLOS:
		DS = ue[ue_idx].LSPs[bs_idx](NLOS_DS);
		ASA = ue[ue_idx].LSPs[bs_idx](NLOS_ASA);
		ASD = ue[ue_idx].LSPs[bs_idx](NLOS_ASD);
		ZSA = ue[ue_idx].LSPs[bs_idx](NLOS_ZSA);
		ZSD = ue[ue_idx].LSPs[bs_idx](NLOS_ZSD);
		mu_XPR = XPR_mean[NLOS];
		sigma_XPR = XRP_std[NLOS];
		break;
	case OUT2IN:
		DS = ue[ue_idx].LSPs[bs_idx](OUT2IN);
		ASA = ue[ue_idx].LSPs[bs_idx](OUT2IN_ASA);
		ASD = ue[ue_idx].LSPs[bs_idx](OUT2IN_ASD);
		ZSA = ue[ue_idx].LSPs[bs_idx](OUT2IN_ZSA);
		ZSD = ue[ue_idx].LSPs[bs_idx](OUT2IN_ZSD);
		mu_XPR = XPR_mean[OUT2IN];
		sigma_XPR = XRP_std[OUT2IN];
		break;
	}
	if (pathloss_model == UMa_A || pathloss_model == UMa_B || pathloss_model == UMi_A || pathloss_model == UMi_B)
	{
		if (LOS_condition == LOS)
		{
			ZSD = ue[ue_idx].LSPs[bs_idx](LOS_ZSD);
		}
		else
		{
			ZSD = ue[ue_idx].LSPs[bs_idx](NLOS_ZSD);
		}
		
	}
	
	Generate_delays(K, DS, tau,tau_LOS, N, propagation_condition);

	N = Generate_cluster_powers(K_R, P_1LOS, DS,tau, P, N, propagation_condition);

	Generate_arrival_departure_angle( K, P, ASA, ASD, ZSA, ZSD, mu_offset_ZOD, N, M, propagation_condition, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, phi_LOS_AOA, phi_LOS_AOD, theta_LOS_ZOA, theta_LOS_ZOD);

	Couple_rays();

	Generate_XPR(mu_XPR,sigma_XPR, N, M, kappa);
	
	
	Compute_Channel_Coef(H,P, K_R, N, M, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, phi_LOS_AOA, phi_LOS_AOD, theta_LOS_ZOA, theta_LOS_ZOD, kappa, LOS_condition, v);



	
	if (check_delay == true)
	{
		cout << endl;
		cout << "**********    delay    **********" << endl;
		cout << endl;

		for (int n = 0; n < N; n++)
		{
			cout << "tau[" << n << "] = " << tau[n] << endl;
		}

		if (propagation_condition == LOS)
		{
			for (int n = 0; n < N; n++)
			{
				cout << "tau_LOS[" << n << "] = " << tau_LOS[n] << endl;
			}
		}
		getchar();
	}
	if (check_cluster_power == true)
	{
		cout << endl;
		cout << "**********  cluster_power ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			cout << "P[" << n << "] = " << P[n] << endl;
		}
		getchar();
	}
	if (check_AOA == true)
	{
		cout << endl;
		cout << "**********  AOA ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				cout << "phi_AOA[" << n << "][" << m << "] : " << phi_AOA[n][m] << endl;
			}
			if (n == 1)
			{
				getchar();
			}
		}
		getchar();
	}
	if (check_AOD == true)
	{
		cout << endl;
		cout << "**********  AOD ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				cout << "phi_AOD[" << n << "][" << m << "] : " << phi_AOD[n][m] << endl;
			}
			if (n == 1)
			{
				getchar();
			}
		}
		getchar();
	}
	if (check_ZOA == true)
	{
		cout << endl;
		cout << "**********  ZOA ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				cout << "theta_ZOA[" << n << "][" << m << "] : " << theta_ZOA[n][m] << endl;
			}
			if (n == 1)
			{
				getchar();
			}
		}
		getchar();
	}
	if (check_ZOD == true)
	{
		cout << endl;
		cout << "**********  ZOD ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				cout << "theta_ZOD[" << n << "][" << m << "] : " << theta_ZOD[n][m] << endl;
			}
			if (n == 1)
			{
				getchar();
			}
		}
		getchar();
	}
	if (check_XPR == true)
	{
		cout << endl;
		cout << "**********  XPR ***********" << endl;
		cout << endl;
		for (int n = 0; n < N; n++)
		{
			for (int m = 0; m < M; m++)
			{
				cout << "XPR[" << n << "][" << m << "] : " << kappa[n][m] << endl;
			}
		}
		getchar();
	}
	
	for (int n = 0; n < N; n++)
	{
		delete[] phi_AOA[n];
		delete[] phi_AOD[n];
		delete[] theta_ZOA[n];
		delete[] theta_ZOD[n];
		delete[] kappa[n];
	}
	
	delete[] phi_AOA;
	delete[] phi_AOD;
	delete[] theta_ZOA;
	delete[] theta_ZOD;
	delete[] tau;
	delete[] P;
	delete[] kappa;
	delete[] H;
}

void Generate_delays(double K,double DS, double *tau,double *tau_LOS, int num_cluster, int propagation_condition)
{
	double X;
	double C_tau;
	double r_tau = delay_scaling_param[propagation_condition];
	double *delay_prime;
	delay_prime = new double[num_cluster];

	for (int n = 0; n < num_cluster; n++)
	{
		if (calibration == true)
		{
			X = log(0.1 + n * 0.05);
		}
		else
		{
			X = log(Rand.u());
		}
		
		delay_prime[n] = - r_tau * DS *X;  
	}
	
	sort(delay_prime, delay_prime + num_cluster);
		
	for (int n = 0; n < num_cluster; n++)
	{
		tau[n] = delay_prime[n] - delay_prime[0];
	
	}


	if (propagation_condition == LOS)
	{
		C_tau = 0.7705 - 0.0433*K + 0.0002*pow(K, 2) + 0.000017*pow(K, 3);
		
		for (int n = 0; n < num_cluster; n++)
		{
			tau_LOS[n] = tau[n]/C_tau ;
		}
	}

	delete[] delay_prime;
}

int Generate_cluster_powers(double K_R, double P_1LOS, double DS, double *tau, double *P, int num_cluster, int propagation_condition)
{
	double *P_prime;
	double Z;
	double r_tau = delay_scaling_param[propagation_condition];
	double zeta = per_cluster_showing_std[propagation_condition];
	double sum_P_prime = 0;
	P_prime = new double[num_cluster];

	for (int n = 0; n < num_cluster; n++)
	{
		if (calibration == true)
		{
			Z = zeta;
		}
		else
		{
			Z = zeta * Rand.u();;
		}
		
		P_prime[n] = exp(-tau[n] * (r_tau - 1) / (r_tau*DS)) *pow(10, -Z / 10);
		sum_P_prime += P_prime[n];
	}


	for (int n = 0; n < num_cluster; n++)
	{
		P[n] = P_prime[n] / sum_P_prime;	
	}


	
	if (propagation_condition == LOS)
	{
		for (int n = 0; n < num_cluster; n++)
		{
			P[n] = 1/(K_R+1) * P_prime[n]/sum_P_prime;
			if (n == 0)
			{
				P[n] += P_1LOS;
			}
		}
	}



	delete[] P_prime;


	double power_treshold = pow(10,-2.5); // -25dB

	/*
	for (int n = 1; n < num_cluster; n++)
	{
		if (P[n]/P[0]<power_treshold)
		{
			return n + 1;
		}
	}
	*/
	return num_cluster;
}

void Generate_arrival_departure_angle(double K, double *P, double ASA, double ASD, double ZSA, double ZSD,  double mu_offset_ZOD, int N, int M, int propagation_condition, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD)
{
	double C_phi;
	double C_phi_NLOS;
	double C_theta;
	double C_theta_NLOS;
	double X_AOA;
	double X_AOD;
	double X_ZOA;
	double X_ZOD;
	double X_0_AOA;
	double X_0_AOD;
	double X_0_ZOA;
	double X_0_ZOD;
	double Y_AOA;
	double Y_AOD;
	double Y_ZOA;
	double Y_ZOD;
	double Y_0_AOA;
	double Y_0_AOD;
	double Y_0_ZOA;
	double Y_0_ZOD;
	double alpha;
	double c_ASA = C_ASA[propagation_condition];
	double c_ASD = C_ASD[propagation_condition];
	double c_ZSA = C_ZSA[propagation_condition];
	double *phi_n_AOA;
	double *phi_n_AOD;
	double *phi_prime_AOA;
	double *phi_prime_AOD;
	double *theta_n_ZOA;
	double *theta_n_ZOD;
	double *theta_prime_ZOA;
	double *theta_prime_ZOD;


	phi_prime_AOA = new double[N];
	phi_prime_AOD = new double[N];
	phi_n_AOA = new double[N];
	phi_n_AOD = new double[N];

	theta_prime_ZOA = new double[N];
	theta_prime_ZOD = new double[N];
	theta_n_ZOA = new double[N];
	theta_n_ZOD = new double[N];

	double P_max;
	P_max = max(P, N);


	if (AOD_AOA_distribution == Wrapped_Gaussian)
	{
		switch (num_cluster[propagation_condition])
		{
		case 4:
			C_phi_NLOS = 0.779;
			break;
		case 5:
			C_phi_NLOS = 0.860;
			break;
		case 8:
			C_phi_NLOS = 1.018;
			break;
		case 10:
			C_phi_NLOS = 1.090;
			break;
		case 11:
			C_phi_NLOS = 1.123;
			break;
		case 12:
			C_phi_NLOS = 1.146;
			break;
		case 14:
			C_phi_NLOS = 1.190;
			break;
		case 15:
			C_phi_NLOS = 1.211;
			break;
		case 16:
			C_phi_NLOS = 1.226;
			break;
		case 19:
			C_phi_NLOS = 1.273;
			break;
		case 20:
			C_phi_NLOS = 1.289;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}
		
		if (propagation_condition == LOS)
		{
			C_phi = C_phi_NLOS * (1.1035 - 0.028*K - 0.002*pow(K, 2) + 0.0001*pow(K, 3));
		}
		else
		{
			C_phi = C_phi_NLOS;
		}

	
		for (int n = 0; n < N; n++)
		{
			phi_prime_AOA[n] = 2 * (ASA / 1.4)*pow(-log(P[n] / P_max), 0.5) / C_phi;
			phi_prime_AOD[n] = 2 * (ASD / 1.4)*pow(-log(P[n] / P_max), 0.5) / C_phi;

		}
	}
	else if(AOD_AOA_distribution == Laplacian)
	{
		switch (num_cluster[propagation_condition])
		{
		case 15:
			C_phi_NLOS = 1.434;
			break;
		case 19:
			C_phi_NLOS = 1.501;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}

		if (propagation_condition == LOS)
		{
			C_phi = C_phi_NLOS * (0.9275 + 0.0439*K - 0.0071*pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_phi = C_phi_NLOS;
		}

		for (int n = 0; n < N; n++)
		{
			phi_prime_AOA[n] = -ASA * log(P[n] / P_max) / C_phi;
			phi_prime_AOD[n] = -ASD * log(P[n] / P_max) / C_phi;

		}

	}
	else
	{
		cout << "check void Generate_arrival_departure_angle()" << endl;
		getchar();
	}



	for (int n = 0 ; n < N; n++)
	{
		double a;
		if (calibration == true)
		{
			a = 0;
		}
		else
		{
			a = Rand.u();
		}

		if (a > 0.5)
		{
			X_AOA = -1;
			if (n == 0)
			{
				X_0_AOA = X_AOA;
			}
		}
		else
		{
			X_AOA = 1;
			if (n == 0)
			{
				X_0_AOA = X_AOA;
			}
		}
		
		if (calibration == true)
		{
			a = 0;
		}
		else
		{
			a = Rand.u();
		}

		if (a > 0.5)
		{
			X_AOD = -1;
			if (n == 0)
			{
				X_0_AOD = X_AOD;
			}
		}
		else
		{
			X_AOD = 1;
			if (n == 0)
			{
				X_0_AOD = X_AOD;
			}
		}

		if (calibration == true)
		{
			Y_AOA = ASA / 7; 
			Y_AOD = ASD / 7;
		}
		else
		{
			Y_AOA = ASA / 7*Rand.n();
			Y_AOD = ASD / 7 * Rand.n();
		}
			
		if (n == 0)
		{
			Y_0_AOA = Y_AOA;
			Y_0_AOD = Y_AOD;
		}
		
		/*
		cout << "Y_AOA : " << Y_AOA << endl;
		cout << "Y_AOD : " << Y_AOD << endl;
		getchar();
		*/


		if (calibration == true)
		{
			phi_LOS_AOA = 0;
			phi_LOS_AOD = 0;
		}


		if (propagation_condition == LOS)
		{
			if (n == 0)
			{
				phi_n_AOA[0] = phi_LOS_AOA;
				phi_n_AOD[0] = phi_LOS_AOD;
			}
			else
			{
				phi_n_AOA[n] = (X_AOA * phi_prime_AOA[n] + Y_AOA) - (X_0_AOA*phi_prime_AOA[0] + Y_0_AOA - phi_LOS_AOA);
				phi_n_AOD[n] = (X_AOD * phi_prime_AOD[n] + Y_AOD) - (X_0_AOD*phi_prime_AOD[0] + Y_0_AOD - phi_LOS_AOD);
			}
			
		}
		else
		{
			phi_n_AOA[n] = X_AOA * phi_prime_AOA[n] + Y_AOA + phi_LOS_AOA;
			phi_n_AOD[n] = X_AOD * phi_prime_AOD[n] + Y_AOD + phi_LOS_AOD;
		}

		
		for (int m = 0; m < M; m++)
		{
			if (m < 2)
			{
				alpha = 0.0447;
			}
			else if (m<4)
			{
				alpha = 0.1413;
			}
			else if (m<6)
			{
				alpha = 0.2492;
			}
			else if (m<8)
			{
				alpha = 0.3715;
			}
			else if (m<10)
			{
				alpha = 0.5129;
			}
			else if (m<12)
			{
				alpha = 0.6797;
			}
			else if (m<14)
			{
				alpha = 0.8844;
			}
			else if (m<16)
			{
				alpha = 1.1481;
			}
			else if (m<18)
			{
				alpha = 1.5195;
			}
			else if (m<20)
			{
				alpha = 2.1551;
			}
			else
			{
				cout << "check num_rays_per_cluster" << endl;
				getchar();
			}

			if ( (m%2) == 1 )
			{
				alpha = -alpha;
			}
			phi_AOA[n][m] = phi_n_AOA[n] + c_ASA * alpha;
			phi_AOD[n][m] = phi_n_AOD[n] + c_ASD * alpha;

		}

	}





	////////////////////////////////// generate ZOA, ZOD//////////////////////////////////////////
	if (pathloss_model == InH_A || pathloss_model == UMa_A || pathloss_model == UMi_A || pathloss_model == RMa_A)
	{
		switch (num_cluster[propagation_condition])
		{
		case 10:
			C_theta_NLOS = 0.9854;
			break;
		case 11:
			C_theta_NLOS = 1.013;
			break;
		case 12:
			C_theta_NLOS = 1.04;
			break;
		case 15:
			C_theta_NLOS = 1.1088;
			break;
		case 19:
			C_theta_NLOS = 1.1764;
			break;
		case 20:
			C_theta_NLOS = 1.1918;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}

		if (propagation_condition == LOS)
		{
			C_theta = C_theta_NLOS * (1.35 + 0.0202*K - 0.0077*pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_theta = C_theta_NLOS;
		}

	}
	else if(pathloss_model == InH_B || pathloss_model == UMa_B || pathloss_model == UMi_B || pathloss_model == RMa_B)
	{
		switch (num_cluster[propagation_condition])
		{
		case 8:
			C_theta_NLOS = 0.889;
			break;
		case 10:
			C_theta_NLOS = 0.957;
			break;
		case 11:
			C_theta_NLOS = 1.031;
			break;
		case 12:
			C_theta_NLOS = 1.104;
			break;
		case 15:
			C_theta_NLOS = 1.1088;
			break;
		case 19:
			C_theta_NLOS = 1.184;
			break;
		case 20:
			C_theta_NLOS = 1.178;
			break;
		default:
			cout << "check num_cluster" << endl;
			break;
		}
		if (propagation_condition == LOS)
		{
			C_theta = C_theta_NLOS * (1.3086 + 0.0339*K - 0.0077 *pow(K, 2) + 0.0002*pow(K, 3));
		}
		else
		{
			C_theta = C_theta_NLOS;
		}
	}
	else
	{
		cout << "check pathloss_model" << endl;
		getchar();
	}
	
		
	for (int n = 0; n < N; n++)
	{
		theta_prime_ZOA[n] = -ZSA * log(P[n] / P_max) / C_theta;
		theta_prime_ZOD[n] = -ZSD * log(P[n] / P_max) / C_theta;
	}
	
	for (int n = 0; n < N; n++)
	{
		if (Rand.u() < 0.5)
		{
			X_ZOA = -1.;
			if (n == 0)
			{
				X_0_ZOA = X_ZOA;
			}
		}
		else
		{
			X_ZOA = 1.;
			if (n == 0)
			{
				X_0_ZOA = X_ZOA;
			}
		}

		if (Rand.u() < 0.5)
		{
			X_ZOD = -1.;
			if (n == 0)
			{
				X_0_ZOD = X_ZOD;
			}
		}
		else
		{
			X_ZOD = 1.;
			if (n == 0)
			{
				X_0_ZOD = X_ZOD;
			}
		}

		if (calibration == true)
		{
			X_ZOA = 1.;
			X_0_ZOA = 1.;
			X_ZOD = 1.;
			X_0_ZOD = 1.;
			theta_LOS_ZOA = 0;
			theta_LOS_ZOD = 0;
			Y_ZOA = ZSA / 7.; 
			Y_ZOD = ZSD / 7.; 
		}
		else
		{
			Y_ZOA = ZSA / 7. * Rand.n(); 
			Y_ZOD = ZSD / 7. * Rand.n();
		}

		if (n == 0)
		{
			Y_0_ZOA = Y_ZOA;
			Y_0_ZOD = Y_ZOD;
		}


		if (propagation_condition == LOS)
		{
			if (n == 0)
			{
				theta_n_ZOA[0] = theta_LOS_ZOA;
				theta_n_ZOD[0] = theta_LOS_ZOD;
			}
			else
			{
				theta_n_ZOA[n] = (X_ZOA * theta_prime_ZOA[n] + Y_ZOA) - (X_0_ZOA*theta_prime_ZOA[0] + Y_0_ZOA - theta_LOS_ZOA);
				theta_n_ZOD[n] = (X_ZOD * theta_prime_ZOD[n] + Y_ZOD) - (X_0_ZOD*theta_prime_ZOD[0] + Y_0_ZOD - theta_LOS_ZOD);
			}

		}
		else
		{
			if (propagation_condition == OUT2IN)
			{
				theta_n_ZOA[n] = X_ZOA * theta_prime_ZOA[n] + Y_ZOA + 90;
			}
			else
			{
				theta_n_ZOA[n] = X_ZOA * theta_prime_ZOA[n] + Y_ZOA + theta_LOS_ZOA;
			}
			
			theta_n_ZOD[n] = X_ZOD * theta_prime_ZOD[n] + Y_ZOD + theta_LOS_ZOD + mu_offset_ZOD;
		}

		

		
		for (int m = 0; m < M; m++)
		{
			if (m < 2)
			{
				alpha = 0.0447;
			}
			else if (m<4)
			{
				alpha = 0.1413;
			}
			else if (m<6)
			{
				alpha = 0.2492;
			}
			else if (m<8)
			{
				alpha = 0.3715;
			}
			else if (m<10)
			{
				alpha = 0.5129;
			}
			else if (m<12)
			{
				alpha = 0.6797;
			}
			else if (m<14)
			{
				alpha = 0.8844;
			}
			else if (m<16)
			{
				alpha = 1.1481;
			}
			else if (m<18)
			{
				alpha = 1.5195;
			}
			else if (m<20)
			{
				alpha = 2.1551;
			}
			else
			{
				cout << "check num_rays_per_cluster" << endl;
				getchar();
			}

			if ((m % 2) == 1)
			{
				alpha = -alpha;
			}
			
			theta_ZOA[n][m] = theta_n_ZOA[n] + c_ZSA * alpha;
			theta_ZOD[n][m] = theta_n_ZOD[n] + 3./8.*pow(10,log_ZSD_mean[propagation_condition])*alpha;
		
		}
	}

	/*
	for (int n = 0; n < N; n++)
	{
	cout<< "phi_prime_AOA["<<n<<"] = " << phi_prime_AOA[n] << endl;
	}
	getchar();

	for (int n = 0; n < N; n++)
	{
	cout << "phi_prime_AOD[" << n << "] = " << phi_prime_AOD[n] << endl;
	}
	getchar();
	*/

	/*
	///// * for calibration   
	cout << endl;
	cout << "******  cluster_AOA  **********" << endl;
	for (int n = 0; n < N; n++)
	{
		cout << "phi_n_AOA["<< n<<"] : " << phi_n_AOA[n] << endl;
	}
	getchar();

	cout << endl;
	cout << "******  cluster_AOD  **********" << endl;
	for (int n = 0; n < N; n++)
	{
		cout << "phi_n_AOD[" << n << "] : " << phi_n_AOD[n] << endl;
	}
	getchar();
	///// * for calibration /// 
	*/

	/*
	///// * for calibration   
	cout << endl;
	cout << "******  cluster_ZOA  **********" << endl;
	for (int n = 0; n < N; n++)
	{
	cout << "theta_n_ZOA["<< n<<"] : " << theta_n_ZOA[n] << endl;
	}
	getchar();

	cout << endl;
	cout << "******  cluster_ZOD  **********" << endl;
	for (int n = 0; n < N; n++)
	{
	cout << "theta_n_ZOD[" << n << "] : " << theta_n_ZOD[n] << endl;
	}
	getchar();
	///// * for calibration ///
	*/



	delete[] phi_n_AOA;
	delete[] phi_n_AOD;
	delete[] phi_prime_AOD;
	delete[] phi_prime_AOA;
	delete[] theta_n_ZOA;
	delete[] theta_n_ZOD;
	delete[] theta_prime_ZOA;
	delete[] theta_prime_ZOD;
	
}

void Couple_rays()
{

}
void Generate_XPR(double mu_XPR, double sigma_XPR, int N, int M, double **kappa)
{
	for (int n=0; n<N; n++)
	{
		for (int m = 0; m<M; m++)
		{
			if (calibration == true)
			{
				kappa[n][m] = pow(10., (sigma_XPR + mu_XPR) / 10.);
			}
			else
			{
				kappa[n][m] = pow(10., (sigma_XPR*Rand.n() + mu_XPR) / 10.);
			}
		}
	}

}



/*===================================================================
FUNCTION: Compute_Channel_Coef()

AUTHOR: 

DESCRIPTION: Compute channel coefficients.

NOTES: Output is a Channel_Coef.

===================================================================*/

void Compute_Channel_Coef(MatrixXcd *H, double *P,double K_R, int N, int M, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD, double **kappa, int LOS_condition, double *v)
{

	double F_rx_theta = 1;
	double F_rx_phi = 1;
	double F_tx_theta = 1;
	double F_tx_phi = 1;

	MatrixXcd *H_NLOS;
	MatrixXcd H_LOS_1;
	
	H_NLOS = new MatrixXcd[N + 4];

	for (int n=0; n<N+4;n++)
	{
		H_NLOS[n] = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, P_ue*Mg_ue*Ng_ue*M_ue*N_ue);
		
	}
	
	H_LOS_1 = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, P_ue*Mg_ue*Ng_ue*M_ue*N_ue);


	
	
	for (int n=0 ; n < N+4; n++)
	{
		if (n < 6)
		{
			int k;
			if (n <3 )
			{
				k = 0;
			}
			else
			{
				k = 1;
			}

			if (n == 0 || n == 3)
			{
				for (int m = 0; m < M; m++)
				{
					if (m <8 || m==18 || m==19)
					{
						H_NLOS[n] += compute_H_NLOS(F_rx_theta, F_rx_phi, F_tx_theta, F_tx_phi, P, k, m, M, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, kappa, v);
					}
				}
			}
			else if (n == 1 || n == 4)
			{
				for (int m = 0; m < M; m++)
				{
					if (m==8 || m == 9 || m == 10 || m == 11 || m == 16 || m == 17)
					{
						H_NLOS[n] += compute_H_NLOS(F_rx_theta, F_rx_phi, F_tx_theta, F_tx_phi, P, k, m, M, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, kappa, v);
					}
				}
			}
			else if (n == 2 || n == 5)
			{
				for (int m = 0; m < M; m++)
				{
					if (m == 12 || m == 13 || m == 14 || m == 15 )
					{
						H_NLOS[n] += compute_H_NLOS(F_rx_theta, F_rx_phi, F_tx_theta, F_tx_phi, P, k, m, M, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, kappa, v);
					}
				}
			}
			else
			{ 
				cout << "check generateSSP.cpp void Compute_Channel_Coef" << endl;
				getchar();
			}
			

		}
		else
		{
			int k = n - 4;
			for (int m = 0; m < M; m++)
			{
				H_NLOS[n] += compute_H_NLOS(F_rx_theta, F_rx_phi, F_tx_theta, F_tx_phi, P, k, m, M, phi_AOA, phi_AOD, theta_ZOA, theta_ZOD, kappa, v);
			}

		}
	}
	
	
	if (LOS_condition == LOS)
	{

		H_LOS_1 = compute_H_LOS(F_rx_theta, F_rx_phi, F_tx_theta, F_tx_phi, phi_LOS_AOA, phi_LOS_AOD, theta_LOS_ZOA, theta_LOS_ZOD, v);
		
		for (int n = 0; n < N + 4; n++)
		{
			if (n == 0)
			{
				H[n] = pow(K_R + 1.0, 0.5)* H_NLOS[n] + pow(K_R+1.0,0.5) * H_LOS_1;
			}
			else
			{
				H[n] = pow(K_R+1,0.5)* H_NLOS[n];
			}
		}
	}
	else
	{
		for (int n = 0; n < N + 4; n++)
		{
			H[n] = H_NLOS[n];
		}
	}

	delete[] H_NLOS;
}


MatrixXcd compute_H_NLOS(double F_rx_theta, double F_rx_phi, double F_tx_theta, double F_tx_phi, double *P, int n, int m,int M, double **phi_AOA, double **phi_AOD, double **theta_ZOA, double **theta_ZOD, double **kappa, double *v)
{
	MatrixXcd H_NLOS;
	MatrixXd r_rx;
	MatrixXd r_tx;
	MatrixXcd a_rx;
	MatrixXcd a_tx;
	complex <double> j = (0, 1);


	H_NLOS = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, P_ue*Mg_ue*Ng_ue*M_ue*N_ue);

	r_rx = MatrixXd(3, 1);
	r_tx = MatrixXd(3, 1);
	a_rx = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, 1);
	a_tx = MatrixXcd::Zero(P_ue*Mg_ue*Ng_ue*M_ue*N_ue, 1);
	

	
	complex <double> jPHI_theta_theta(0, 2.0 * PI *Rand.u() - PI);
	complex <double> jPHI_theta_phi(0, 2.0 * PI *Rand.u() - PI);
	complex <double> jPHI_phi_theta(0, 2.0 * PI *Rand.u() - PI);
	complex <double> jPHI_phi_phi(0, 2.0 * PI *Rand.u() - PI);

	r_rx(0,0) = sin(theta_ZOA[n][m])*cos(phi_AOA[n][m]);
	r_rx(1,0) = sin(theta_ZOA[n][m])*sin(phi_AOA[n][m]);
	r_rx(2,0) = cos(theta_ZOA[n][m]);

	r_tx(0,0) = sin(theta_ZOD[n][m])*cos(phi_AOD[n][m]);
	r_tx(1,0) = sin(theta_ZOD[n][m])*sin(phi_AOD[n][m]);
	r_tx(2,0) = cos(theta_ZOD[n][m]);

	double r_rx_v_t = t * (r_rx(0, 0) * v[0] + r_rx(1, 0) * v[1] + r_rx(2, 0) * v[2]);


	a_rx = j * 2.0*PI / lamda_0 * W_rx * r_rx;
	a_tx = j * 2.0*PI / lamda_0 * W_tx * r_tx;

	H_NLOS = pow(P[n] / M, 0.5) *(F_tx_theta * (F_rx_theta * exp(jPHI_theta_theta) + F_rx_phi * pow(1.0 / kappa[n][m], 0.5) * exp(jPHI_phi_theta))
			+ F_tx_phi * (F_rx_theta * pow(1.0 / kappa[n][m], 0.5) * exp(jPHI_theta_phi) + F_rx_phi * exp(jPHI_phi_phi))) *
		a_rx.array().exp().matrix()*a_tx.array().exp().matrix().adjoint()*exp(j*2.0*PI*r_rx_v_t/ lamda_0);
	
	return H_NLOS;
}

MatrixXcd compute_H_LOS(double F_rx_theta, double F_rx_phi, double F_tx_theta, double F_tx_phi, double phi_LOS_AOA, double phi_LOS_AOD, double theta_LOS_ZOA, double theta_LOS_ZOD, double *v)
{
	MatrixXcd H_LOS;
	MatrixXd r_rx;
	MatrixXd r_tx;

	MatrixXcd a_rx;
	MatrixXcd a_tx;
	complex <double> j = (0, 1);
	

	H_LOS = MatrixXcd::Zero(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, P_ue*Mg_ue*Ng_ue*M_ue*N_ue);
	
	r_rx = MatrixXd(3, 1);
	r_tx = MatrixXd(3, 1);
	a_rx = MatrixXd(P_bs*Mg_bs*Ng_bs*M_bs*N_bs, 1);
	a_tx = MatrixXd(P_ue*Mg_ue*Ng_ue*M_ue*N_ue, 1);

	
	complex <double> jPHI_LOS(0, 2.0 * PI *Rand.u() - PI);

	
	r_rx(0, 0) = sin(theta_LOS_ZOA)*cos(phi_LOS_AOA);
	r_rx(1, 0) = sin(theta_LOS_ZOA)*sin(phi_LOS_AOA);
	r_rx(2, 0) = cos(theta_LOS_ZOA);

	r_tx(0, 0) = sin(theta_LOS_ZOD)*cos(phi_LOS_AOD);
	r_tx(1, 0) = sin(theta_LOS_ZOD)*sin(phi_LOS_AOD);
	r_tx(2, 0) = cos(theta_LOS_ZOD);

	a_rx = j * 2.0*PI / lamda_0 * W_rx * r_rx;
	a_tx = j * 2.0*PI / lamda_0 * W_tx * r_tx;

	double r_rx_v_t = t * (r_rx(0, 0) * v[0] + r_rx(1, 0) * v[1] + r_rx(2, 0) * v[2]);
	complex <double> j2_PI_r_rx_v_t(0, 2 * PI * r_rx_v_t);
	
	
	H_LOS = (F_rx_theta * F_tx_theta - F_rx_theta*F_rx_phi) * exp(jPHI_LOS) * a_rx.array().exp().matrix()*a_tx.array().exp().matrix().adjoint() * exp(j2_PI_r_rx_v_t / lamda_0);
	
	return H_LOS;
}

double max(double *A, int size)
{
	double max_value=0;
	for (int n = 0; n< size; n++)
	{
		if (n == 0)
		{
			max_value = A[0];
		}
		else 
		{
			if (A[n] > max_value)
			{
				max_value = A[n];
			}
		}
	}
	return max_value;
}
