#include "ue.h"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include "const.h"
#include "struct.h"
#include "paramSimul.h"
#include "paramFastFading.h"
#include "evaluationConfigurerations.h"
using namespace std;

/*===================================================================
FUNCTION: UE::UE()

AUTHOR: SJ

DESCRIPTION: UE Object Constructor

NOTES: Currently, nothing is done when a UE is instantiated

===================================================================*/
UE::UE()
{
	if (test_environment > 2)
	{
		pos = new LOCATION[19];
	}

	_2D_distance = new double[num_bs];
	break_point_distance = new double[num_bs];
	pathloss = new double[num_bs];
	LOS_condition = new int[num_bs];
	propagation_condition = new int[num_bs];
	TLSPs = new ArrayXd[num_bs];
	LSPs = new ArrayXd[num_bs];
	ZOD_offset = new double[num_bs];
	antennagain = new double[num_sector];

	phi_LOS_AOD = new double[num_sector];
	phi_LOS_AOA = new double[num_sector];
	theta_LOS_ZOD = new double[num_sector];
	theta_LOS_ZOA = new double[num_sector];
	
	ue2sector_azimuth_angle = new double[num_sector];
	LS_gain = new double[num_sector];
	adj_sector = new int[num_compute_coef];
	sigma_SF = new double[num_bs];
}

/*===================================================================
FUNCTION: UE::~UE()

AUTHOR: SJ

DESCRIPTION: UE Object Destructor

NOTES: Currently, nothing is done

===================================================================*/
UE::~UE()
{
	delete[] _2D_distance;
	delete[] ue2sector_azimuth_angle;
	delete[] LOS_condition;
	delete[] pathloss;
	delete[] antennagain;
	delete[] LS_gain;
	delete[] break_point_distance;
	delete[] adj_sector;
	delete[] pos;
	delete[] TLSPs;
	delete[] LSPs;
	delete[] propagation_condition;
	delete[] phi_LOS_AOD;
	delete[] phi_LOS_AOA;
	delete[] theta_LOS_ZOD;
	delete[] theta_LOS_ZOA;
	delete[] ZOD_offset;
	delete[] sigma_SF;
}


/*===================================================================
FUNCTION: UE::Configure( int ue_idx )

AUTHOR: JH

DESCRIPTION: UE object gets to know the index of itself

NOTES:

===================================================================*/
void UE::Configure(int ue_idx)
{
	Set_UE_speed_indoor_outdoor_state();
	Set_h_UT();
	_ue_idx = ue_idx;
	Pathloss_car = 0;
	Pathloss_tw = 0;

}


void UE::Allocate_memory()
{


}


void UE::Set_UE_speed_indoor_outdoor_state()
{
	if (test_environment <3)
	{
		speed = indoor_ue_speed * 1000. / 3600.;
	}
	
	else if (test_environment<9)
	{
		if (Rand.u() < indoor_ue_portion)
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}		
	}
	else
	{
		if (Rand.u() < indoor_ue_portion)
		{
			ue_indoor_outdoor_state = INDOOR;
			
			speed = indoor_ue_speed * 1000. / 3600.;
		}
		else if(Rand.u()< outdoor_pedestrian_portion/(1- indoor_ue_portion))
		{
			ue_indoor_outdoor_state = PEDESTRIAN;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
		else
		{
			ue_indoor_outdoor_state = IN_CAR;
			speed = outdoor_ue_car_speed * 1000. / 3600.;
		}
	}

	double theta_v = 2 * PI*Rand.u();
	double phi_v = 2 * PI*Rand.u();
	
	v[0] = speed * sin(theta_v)*cos(phi_v);
	v[1] = speed * sin(theta_v)*sin(phi_v);
	v[2] = speed * cos(theta_v);

	
	if (ue_indoor_outdoor_state == INDOOR)
	{
		for (int bs_idx=0; bs_idx<num_bs; bs_idx++)
		{
			propagation_condition[bs_idx] = OUT2IN;
		}
		Set_2D_distance_indoor();
		Set_O2I_penet_loss();
	}
	else if (ue_indoor_outdoor_state == IN_CAR)
	{
		Set_car_penet_loss();
	}



}


/*===================================================================
FUNCTION: UE::Set_2D_distance_indoor( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_2D_distance_indoor

NOTES: m.eval2020 page 44,45

===================================================================*/
void UE::Set_2D_distance_indoor()
{	
	if (pathloss_model == UMa_A || pathloss_model == UMi_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 25;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 10);
		}

	}
	else if (pathloss_model==RMa_A)
	{
		if (carrier_freq < 6000000000)
		{
			_2D_distance_indoor = Rand.u() * 10;
		}
		else
		{
			_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 25);
		}
	}
	else if (pathloss_model == UMa_B || pathloss_model == UMi_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 10, Rand.u() * 25);
	}
	else if (pathloss_model == RMa_B)
	{
		_2D_distance_indoor = MIN(Rand.u() * 25, Rand.u() * 10);
	}

}



/*===================================================================
FUNCTION: UE::Set_O2I_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_O2I_penet_loss

NOTES: m.eval2020 page 44,45

===================================================================*/

void UE::Set_O2I_penet_loss()
{	

	if ((pathloss_model == UMa_A|| pathloss_model == UMi_A) && carrier_freq<6000000000)
	{
		Pathloss_tw = 20 + 0.5*_2D_distance_indoor;
	}
	else if (pathloss_model == RMa_A && carrier_freq<6000000000 )
	{
		Pathloss_tw = 10 + 0.5*_2D_distance_indoor;
	}
	else
	{
		if (Rand.u() < high_loss_type_portion)
		{
	
			Pathloss_tw = High_loss_model + 0.5*_2D_distance_indoor + High_loss_model_std*Rand.n();
		}
		else
		{
			Pathloss_tw = Low_loss_model + 0.5*_2D_distance_indoor + Low_loss_model_std*Rand.n();
		}
	}
	
}


/*===================================================================
FUNCTION: UE::Set_car_penet_loss( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_car_penet_loss

NOTES: m.eval2020 page 45

===================================================================*/

void UE::Set_car_penet_loss()
{
	Pathloss_car = car_penet_loss_mean + car_penet_loss_std*Rand.n();
}


/*===================================================================
FUNCTION: UE::Set_h_UT( int ue_idx )

AUTHOR: GJ

DESCRIPTION: Set_h_UT

NOTES: m.eval2020 page 19 Table 8-X1 UE antenna height

===================================================================*/
void UE::Set_h_UT()
{
	double n_fl;
	double N_fl;
	

	if (test_environment>2 && test_environment <7 )
	{
		if (ue_indoor_outdoor_state ==INDOOR)
		{
			N_fl = floor(5.*Rand.u() + 4);
			n_fl = floor(N_fl*Rand.u() + 1);
			h_UT = 3. * (n_fl - 1) + 1.5;
			floor_idx = int(n_fl)-1;

		}
		else
		{
			h_UT = 1.5;
			floor_idx = 0;
		}
	}
	else
	{
		floor_idx = 0;
		h_UT = 1.5;
	}
}


/*===================================================================
FUNCTION: Drop_ue_in_cell()

AUTHOR: G.J

DESCRIPTION: Drop ues

NOTES: UE  sector Ȥ cell ȿ ߸.

OUTPUT : location_x, location_y

===================================================================*/
void UE::Drop_ue_in_cell()
{
	if (test_environment < 3)
	{
		do
		{
			pos_in_cell.x = 20. * Rand.u();
			pos_in_cell.y = 25. * Rand.u();
		} while (sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);

		
	}
	else if (test_environment == Dense_Urban_eMBB_C_Micro)
	{
		do
		{
			pos_in_cell.x = micro_cell_radius * Rand.u();
			pos_in_cell.y = micro_cell_radius * Rand.u();

		} while ((sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance) || (sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) > micro_cell_radius));

		double x = Rand.u();
		if (x < 0.25)
		{
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.5)
		{
			pos_in_cell.x = -pos_in_cell.x;
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.75)
		{
			pos_in_cell.x = -pos_in_cell.x;
		}
	}
	else
	{
		bool a = 0;

		if (a == 0)
		{
			do
			{
				pos_in_cell.x = inter_site_distance / 2.* Rand.u();
				pos_in_cell.y = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.y > (-pos_in_cell.x + inter_site_distance) / sqrt(3.) || sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);
		}
		else
		{
			do
			{
				pos_in_cell.y = inter_site_distance / 2.* Rand.u();
				pos_in_cell.x = inter_site_distance / sqrt(3.)* Rand.u();
			} while (pos_in_cell.x > (-pos_in_cell.y + inter_site_distance) / sqrt(3.)	|| sqrt(pos_in_cell.x*pos_in_cell.x + pos_in_cell.y*pos_in_cell.y) < secter_to_ue_min_distance);

		}

		double x = Rand.u();

		if (x < 0.25)
		{
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.5)
		{
			pos_in_cell.x = -pos_in_cell.x;
			pos_in_cell.y = -pos_in_cell.y;
		}
		else if (x < 0.75)
		{
			pos_in_cell.x = -pos_in_cell.x;
		}

	}
}

/*===================================================================
FUNCTION: Determine_LOS()

AUTHOR: G.J

DESCRIPTION: Determine_LOS

NOTES: m.2020 eval page 46

OUTPUT : ue[ue_idx].los[sec_idx]

===================================================================*/
void UE::Determine_LOS(int bs_idx)
{
	double d_2D =  _2D_distance[bs_idx];
	double d_2D_out = d_2D - _2D_distance_indoor;
	double P_LOS;

	if (test_environment <3)
	{
		if (d_2D < 5)
		{
			P_LOS = 1.0;
		}
		else if(d_2D < 49)
		{
			P_LOS = exp(-(d_2D - 5.) / 70.8);
		}
		else
		{
			P_LOS = exp(-(d_2D - 49.) / 211.7)*0.54;
		}

	}
	else if (test_environment<6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			double C;
				
			if (h_UT <= 13.0)
			{
				C = 0;
			}
			else
			{
				C = pow((h_UT - 13.0) / 10.0, 1.5);
			}

			if (ue_indoor_outdoor_state== INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 63)*(1 - 18 / d_2D_out)) * (1 + C*5.0 / 4.0*pow(d_2D_out / 100.0, 3)*exp(-d_2D_out / 150.0));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D)) * (1 + C*5.0 / 4.0*pow(d_2D / 100.0, 3)*exp(-d_2D / 150.0));
			}

		}
	}
	else if (test_environment == 6)
	{
		if (d_2D <= 18.0)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = (18.0 / d_2D_out + exp(-d_2D_out / 36)*(1 - 18 / d_2D_out));
			}
			else
			{
				P_LOS = (18.0 / d_2D + exp(-d_2D / 63)*(1 - 18 / d_2D));
			}
		}
			

	}
	else
	{
		if (d_2D <= 10.)
		{
			P_LOS = 1.0;
		}
		else
		{
			if (ue_indoor_outdoor_state == INDOOR)
			{
				P_LOS = exp(-(d_2D_out - 10)/1000);
			}
			else
			{
				P_LOS = exp(-(d_2D - 10) / 1000);
			}
		}

	}
		
	if ( Rand.u() < P_LOS)
	{
		LOS_condition[bs_idx] = LOS;
		propagation_condition[bs_idx] = LOS;
	}
	else
	{
		LOS_condition[bs_idx] = NLOS;
		propagation_condition[bs_idx] = NLOS;
	}
}


/*===================================================================
FUNCTION: void UE::Set_breakpoint_distance(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Set_breakpoint_distance

NOTES: 

OUTPUT : ue[ue_idx].break_point_distance[sec_idx]

===================================================================*/
void UE::Set_breakpoint_distance(int bs_idx)
{
	double d_2D = _2D_distance[bs_idx];

	if (test_environment<7)
	{
		double C;
		double g;
		double h_E;


		if (d_2D <= 18.)
		{
			g = 0.;
		}
		else
		{
			g = 5. / 4. * pow(d_2D / 100., 3.)*exp(-d_2D / 150.);
		}

		if (h_UT < 13.)
		{
			C = 0;
		}
		else
		{
			C = pow((h_UT - 13.) / 10., 1.5)*g;
		}


		if (Rand.u() < 1 / (1 + C))
		{
			h_E = 1.;
		}
		else
		{
			h_E = floor((h_UT - 1.5 - 9.) / 3. * Rand.u()) * 3. + 12.;
		}

		break_point_distance[bs_idx] = 4. * (h_BS - h_E) * (h_UT - h_E) * carrier_freq / SPEED_LIGHT;

	}
	else
	{		
		break_point_distance[bs_idx] = 2. * PI * h_BS * h_UT * carrier_freq / SPEED_LIGHT;
	}
}


/*===================================================================
FUNCTION: void UE::Compute_PathLoss(int sec_idx)

AUTHOR: G.J

DESCRIPTION: Compute_PathLoss

NOTES:

OUTPUT : ue[ue_idx].pathloss[bs_idx]

===================================================================*/
void UE::Compute_PathLoss(int bs_idx)
{
	double d_2D = _2D_distance[bs_idx];
	double d_in = _2D_distance_indoor;
	double d_3D = pow(pow(d_2D, 2.) + pow(h_BS - h_UT, 2.), 0.5);
	double d_BP = break_point_distance[bs_idx];
	double fc = carrier_freq / 1000000000.;
	

	if (LOS_condition[bs_idx] == LOS)
	{
		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					pathloss[bs_idx] = 32.8 + 16.9*log10(d_3D) + 20 * log10(fc);

					sigma_SF[bs_idx] = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					pathloss[bs_idx] = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					sigma_SF[bs_idx] = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_A" << endl;
				getchar();
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 &&fc <= 100)
			{
				if (d_2D <= 150)
				{
					pathloss[bs_idx] = 32.4 + 17.3*log10(d_3D) + 20 * log10(fc);
					sigma_SF[bs_idx] = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss InH_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc, LOS Compute_PathLoss pathloss_model InH_B" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
			
					pathloss[bs_idx] = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					pathloss[bs_idx] = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					
					cout << "check d_2D, LOS Compute_PathLoss UMa_A " << endl;
					getchar();
				}

			
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_A" << endl;
				getchar();
			}


		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					pathloss[bs_idx] = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMa_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMa_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 22.*log10(d_3D) + 28.0 + 20.*log10(fc);
					sigma_SF[bs_idx] = 3;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					pathloss[bs_idx] = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					sigma_SF[bs_idx] = 3;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					pathloss[bs_idx] = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_A " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					pathloss[bs_idx] = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model UMi_B" << endl;
				getchar();
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					pathloss[bs_idx] = PL1 + 40 * log10(d_3D / d_BP);
					sigma_SF[bs_idx] = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_A" << endl;
				getchar();
			}

		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					pathloss[bs_idx] = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D;
					sigma_SF[bs_idx] = 4;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					pathloss[bs_idx] = PL1 + 40 * log10(d_3D / d_BP) ;
					sigma_SF[bs_idx] = 6;
				}
				else
				{
					cout << "check d_2D, LOS Compute_PathLoss RMa_B " << endl;
				}
			}
			else
			{
				cout << "check fc , LOS Compute_PathLoss pathloss_model RMa_B" << endl;
				getchar();
			}
		}

	}
	else
	{
		double PL_LOS;
		double PL_NLOS;

		if (pathloss_model == InH_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (d_2D <= 150)
				{
					pathloss[bs_idx] = 43.3*log10(d_3D) + 11.5 + 20.*log10(fc);
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else if (fc >= 6 && fc <= 100)
			{
				if (d_2D <= 150)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
					if(PL_LOS>PL_NLOS)
					{
						sigma_SF[bs_idx] = 3;
					}
					else
					{
						sigma_SF[bs_idx] = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == InH_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (d_2D <= 150.)
				{
					PL_LOS = 32.4 + 17.3*log10(d_3D) + 20.0*log10(fc);
					PL_NLOS = 17.3 + 38.3*log10(d_3D) + 24.9*log10(fc);
					pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
					if (PL_LOS>PL_NLOS)
					{
						sigma_SF[bs_idx] = 3;
					}
					else
					{
						sigma_SF[bs_idx] = 8.03;
					}
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss InH_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss InH_A" << endl;
			}
		}
		else if (pathloss_model == UMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
					
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}

				if (10. <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3)
						+ 20.*log10(fc) - (3.2*pow(log10(17.625), 2) - 4.97) - 0.6*(h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				

				
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					sigma_SF[bs_idx] = 6;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					sigma_SF[bs_idx] = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_A" << endl;
			}

		}
		else if (pathloss_model == UMa_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 28.0 + 22.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.)
				{
					PL_NLOS = 13.54 + 39.08*log10(d_3D) + 20 * log10(fc) - 0.6 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_B " << endl;
				}
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					sigma_SF[bs_idx] = 6;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMa_B" << endl;
			}

		}
		else if (pathloss_model == UMi_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc) ;
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 40.*log10(d_3D) + 28.0 + 20.*log10(fc) - 9.*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 5000.) // !!!!!! Table 3.3 10m<d_2d<2000m ,
				{
					PL_NLOS = 36.7*log10(d_3D) + 22.7 + 26 * log10(fc) - 0.3*(h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMa_A " << endl;
				}
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);

				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 3;
				}
				else
				{
					sigma_SF[bs_idx] = 4;
				}

			}
			else if (fc >= 6 && fc <= 100)
			{
				if (secter_to_ue_min_distance < d_2D && d_2D < d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP < d_2D	&& d_2D < 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}

				if (secter_to_ue_min_distance < d_2D && d_2D < 5000.)   ///!! table3-3 shadowing check in 38.901 7.82
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3 * log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					sigma_SF[bs_idx] = 7.82;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == UMi_B)
		{
			if (fc >= 0.5 && fc <= 100)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 32.4 + 21.*log10(d_3D) + 20.*log10(fc);
				}
				else if (d_BP <= d_2D	&& d_2D <= 5000.)
				{
					PL_LOS = 32.4 + 40 * log10(d_3D) + 20.*log10(fc) - 9.5*log10(pow(d_BP, 2) + pow(h_BS - h_UT, 2));
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_B " << endl;
					getchar();
				}

				if (10. < d_2D && d_2D < 5000.)  ///!! table3-3 shadowing check
				{
					PL_NLOS = 35.3*log10(d_3D) + 22.4 + 21.3* log10(fc) - 0.3 * (h_UT - 1.5);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss UMi_A " << endl;
				}
				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
				if (PL_LOS>PL_NLOS)
				{
					sigma_SF[bs_idx] = 4;
				}
				else
				{
					sigma_SF[bs_idx] = 7.82;
				}

			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss UMi_A" << endl;
			}
		}
		else if (pathloss_model == RMa_A)
		{
			if (fc >= 0.5 && fc <= 6)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					pathloss[bs_idx] = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
					sigma_SF[bs_idx] = 8;
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_A " << endl;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}
		else if (pathloss_model == RMa_B)
		{
			if (fc >= 0.5 && fc <= 30)
			{
				if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
				{
					PL_LOS = 20 * log10(40 * PI*d_3D*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_3D) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_3D ;
				}
				else if (d_BP <= d_2D	&& d_2D <= 21000.)
				{
					double PL1;
					PL1 = 20 * log10(40 * PI*d_BP*fc / 3.) + MIN(0.03*pow(h, 1.72), 10)*log10(d_BP) - MIN(0.044*pow(h, 1.72), 14.77) + 0.002*log10(h)*d_BP;
					PL_LOS = PL1 + 40 * log10(d_3D / d_BP);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				if (secter_to_ue_min_distance <= d_2D && d_2D <= 21000.)
				{
					PL_NLOS = 161.04 - 7.1*log10(W) + 7.5*log10(h) - (24.37 - 3.7*pow((h / h_BS), 2))*log10(h_BS) + (43.42 - 3.1*log10(h_BS))*(log10(d_3D) - 3) + 20 * log10(fc) - (3.2*pow(log10(11.75*h_UT), 2) - 4.97);
				}
				else
				{
					cout << "check d_2D, NLOS Compute_PathLoss RMa_B " << endl;
				}

				pathloss[bs_idx] = MAX(PL_LOS, PL_NLOS);
				
				if (PL_LOS>PL_NLOS)
				{
					if (secter_to_ue_min_distance <= d_2D && d_2D <= d_BP)
					{
						sigma_SF[bs_idx] = 4;
					}
					else
					{
						sigma_SF[bs_idx] = 6;
					}
					
				}
				else
				{
					sigma_SF[bs_idx] = 8;
				}
			}
			else
			{
				cout << "check fc , NLOS Compute_PathLoss RMa_A" << endl;
			}
		}
	}

	if (propagation_condition[bs_idx] == OUT2IN)
	{
		if (pathloss_model == RMa_A || pathloss_model == RMa_B)
		{
			sigma_SF[bs_idx] = 8;
		}
		else
		{
			sigma_SF[bs_idx] = 7;
		}
		
	}

	

	pathloss[bs_idx] += Pathloss_tw + Pathloss_car;

}

void UE::Compute_antenna_gain(int sec_idx)
{
	if (test_environment == InH_eMBB_A || test_environment == InH_eMBB_B || test_environment == InH_eMBB_C) // ceiling mount antenna pattern
	{
		double height = h_BS - h_UT;
		double angle_v = atan2(_2D_distance[sec_idx/(num_sector/num_bs)], height) * 180. / PI + 90;
		double antenna_v = -MIN(12. * pow(((angle_v - 90) / 90), 2), 25);
		//double antenna_h = -MIN(12. * pow((ue2sector_azimuth_angle[sec_idx] / 90.), 2), 25);

		antennagain[sec_idx] = -MIN(-(antenna_v ), 25) + bs_antenna_element_gain;

		//antennagain[sec_idx] = -MIN(-(antenna_v + antenna_h), 25) + bs_antenna_element_gain;

	}
	else
	{
		double height = h_BS - h_UT;
		double angle_v = atan2(height, _2D_distance[sec_idx/(num_sector / num_bs)]) * 180. / PI + 90;
		double antenna_v = -MIN(12. * pow(((angle_v - 90) / 65), 2), 30);
		double antenna_h = -MIN(12. * pow((ue2sector_azimuth_angle[sec_idx] / 65), 2), 30);

		antennagain[sec_idx] = -MIN(-(antenna_v + antenna_h), 30.) + bs_antenna_element_gain;

		// This equation comes from 8.5.1.1 (2) in dB
	}

}